import os

import asdf
import numpy as np
import pytest

_REFFILE_PATH = os.path.join(os.path.dirname(__file__), "..", "reference_files")


def assert_tree_match(tree_a, tree_b):
    seen = set()

    def walk_trees(a, b):
        seen_key = (id(a), id(b))

        if seen_key in seen:
            return

        seen.add(seen_key)

        assert isinstance(a, type(b))
        if isinstance(a, dict):
            assert set(a.keys()) == set(b.keys())
            for k in a:
                walk_trees(a[k], b[k])
        elif isinstance(a, list):
            assert len(a) == len(b)
            for ai, bi in zip(a, b):
                walk_trees(ai, bi)
        elif isinstance(a, (np.ndarray, asdf.tags.core.ndarray.NDArrayType)):
            np.testing.assert_array_equal(a, b)
        else:
            assert a == b

    ignore_keys = {"asdf_library", "history"}
    walk_trees(
        {k: v for k, v in tree_a.items() if k not in ignore_keys},
        {k: v for k, v in tree_b.items() if k not in ignore_keys},
    )


@pytest.mark.parametrize(
    "a, b",
    [
        ({"l": [1, 2, 3]}, {"l": [1, 2, 3]}),
        ({"d": {"a": 1}}, {"d": {"a": 1}}),
        ({"ld": [{"a": 1}]}, {"ld": [{"a": 1}]}),
        ({"ld": [{"a": 1}]}, {"ld": [{"a": 1}]}),
        ({"dk": {"a": 1, "b": 2}}, {"dk": {"b": 2, "a": 1}}),
        ({"a": np.arange(3)}, {"a": np.arange(3)}),
        ({"la": [np.arange(3)]}, {"la": [np.arange(3)]}),
        ({"i": 1}, {"i": 1}),
        ({"s": "abc"}, {"s": "abc"}),
        ({"f": 1.0}, {"f": 1.0}),
        ({"i": 1}, {"i": 1, "asdf_library": None}),
        ({"i": 1}, {"i": 1, "history": None}),
    ],
)
def test_assert_tree_match(a, b):
    assert_tree_match(a, b)


@pytest.mark.parametrize(
    "a, b",
    [
        ({"l": [1, 2, 3]}, {"l": [1, 2]}),
        ({"l": [1, 2, 3]}, {"l": [1, 2, 4]}),
        ({"d": {"a": 1}}, {"d": {"a": 2}}),
        ({"d": {"a": 1}}, {"d": {"b": 1}}),
        ({"d": {"a": 1, "b": 2}}, {"d": {"a": 1}}),
        ({"d": {"a": 1}}, {"d": {"a": 1, "b": 2}}),
        ({"a": np.arange(3)}, {"a": np.arange(4)}),
        ({"la": [np.arange(3)]}, {"a": [np.arange(4)]}),
        ({"i": 1}, {"i": 2}),
        ({"s": "abc"}, {"s": "def"}),
        ({"f": 1.0}, {"f": 1.1}),
    ],
)
def test_assert_tree_match_fail(a, b):
    with pytest.raises(AssertionError):
        assert_tree_match(a, b)


def get_test_id(reference_file_path):
    """Helper function to return the informative part of a schema path"""
    path = os.path.normpath(str(reference_file_path))
    return os.path.sep.join(path.split(os.path.sep)[-3:])


def collect_reference_files():
    """Function used by pytest to collect ASDF reference files for testing."""
    for version in asdf.versioning.supported_versions:
        version_dir = os.path.join(_REFFILE_PATH, str(version))
        if os.path.exists(version_dir):
            for filename in os.listdir(version_dir):
                if filename.endswith(".asdf"):
                    filepath = os.path.join(version_dir, filename)
                    basename, _ = os.path.splitext(filepath)
                    if os.path.exists(basename + ".yaml"):
                        yield filepath


def _compare_trees(name_without_ext):
    asdf_path = name_without_ext + ".asdf"
    yaml_path = name_without_ext + ".yaml"

    with asdf.open(asdf_path) as af_handle:
        af_handle.resolve_references()

        with asdf.open(yaml_path) as ref:
            # TODO strip asdf_library and history
            assert_tree_match(af_handle.tree, ref.tree)


def test_reference_files_exist():
    assert list(collect_reference_files())


@pytest.mark.parametrize("reference_file", collect_reference_files(), ids=get_test_id)
def test_reference_file(reference_file):
    name_without_ext, _ = os.path.splitext(reference_file)
    _compare_trees(name_without_ext)



import os

try:
    import asdf
except ImportError:
    raise ImportError(
        "asdf must be installed to regenerate the reference files.")

import numpy as np


def ref_scalars(fd):
    tree = {
        'int': 42,
        'float': 3.14,
        'string': 'foo',
    }
    asdf.AsdfFile(tree).write_to(fd)


def ref_anchor(fd):
    nestee = {"abc": 123}
    tree = {
        "a": nestee,
        "b": nestee,
    }
    asdf.AsdfFile(tree).write_to(fd)


def ref_endian(fd):
    tree = {
        'big': np.arange(42, dtype=">i4"),
        'little': np.arange(42, dtype="<i4"),
    }

    asdf.AsdfFile(tree).write_to(fd)


def ref_structured(fd):
    tree = {
        'structured': np.array(
            [
                (1, 'a', 3.3),
                (2, 'b', 6.6),
            ], dtype=[
                ('a', np.uint8),
                ('b', 'S3'),
                ('c', np.float32),
            ]),
    }
    asdf.AsdfFile(tree).write_to(fd)

def ref_basic(fd):
    tree = {
        'data': np.arange(8)
        }

    asdf.AsdfFile(tree).write_to(fd)


def ref_int(fd):
    tree = {}

    for size in (1, 2, 4):
        bits = size * 8
        for endian in ['>', '<']:
            values = [(1 << (bits - 1)) - 1, -(1 << (bits - 1)), 0]
            datatype = '%si%d' % (endian, size)
            arr = np.array(values, datatype)
            tree['datatype' + datatype] = arr

            values = [(1 << bits) - 1, 0]
            datatype = '%su%d' % (endian, size)
            arr = np.array(values, datatype)
            tree['datatype' + datatype] = arr

    asdf.AsdfFile(tree).write_to(fd)


def ref_float(fd):
    tree = {}

    for size in (4, 8):
        for endian in ['>', '<']:
            datatype = '%sf%d' % (endian, size)
            finfo = np.finfo(np.dtype(datatype))
            values = [0.0, -0.0, np.nan, np.inf, -np.inf,
                      finfo.min, finfo.max, finfo.eps, finfo.epsneg, finfo.tiny]
            arr = np.array(values, datatype)
            tree['datatype' + datatype] = arr

    asdf.AsdfFile(tree).write_to(fd)


def ref_complex(fd):
    tree = {}

    for size in (4, 8):
        for endian in ['>', '<']:
            datatype = '%sf%d' % (endian, size)
            finfo = np.finfo(np.dtype(datatype))
            values = [0.0, -0.0, np.nan, np.inf, -np.inf,
                      finfo.min, finfo.max, finfo.eps, finfo.epsneg, finfo.tiny]

            complex_values = []
            for x in values:
                for y in values:
                    complex_values.append(x + 1j * y)
            datatype = '%sc%d' % (endian, size * 2)
            arr = np.array(complex_values, datatype)
            tree['datatype' + datatype] = arr

    asdf.AsdfFile(tree).write_to(fd)


def ref_ascii(fd):
    arr = np.array([b'', b'ascii'], dtype='S')
    tree = {'data': arr}
    asdf.AsdfFile(tree).write_to(fd)


def ref_unicode_bmp(fd):
    tree = {}
    for endian in ['>', '<']:
        arr = np.array(['', 'Æʩ'], dtype=endian + 'U')
        tree['datatype' + endian + 'U'] = arr

    asdf.AsdfFile(tree).write_to(fd)


def ref_unicode_spp(fd):
    tree = {}
    for endian in ['>', '<']:
        arr = np.array(['', '𐀠'], dtype=endian + 'U')
        tree['datatype' + endian + 'U'] = arr

    asdf.AsdfFile(tree).write_to(fd)


def ref_shared(fd):
    data = np.arange(8)
    tree = {
        'data': data,
        'subset': data[1::2]
        }

    asdf.AsdfFile(tree).write_to(fd)


def ref_stream(fd):
    tree = {
        # Each "row" of data will have 128 entries.
        'my_stream': asdf.Stream([8], np.float64)
    }

    ff = asdf.AsdfFile(tree)
    with open(fd, 'wb') as fd:
        ff.write_to(fd)
        # Write 100 rows of data, one row at a time.  ``write_to_stream``
        # expects the raw binary bytes, not an array, so we use
        # ``tobytes()``.
        for i in range(8):
            fd.write(np.array([i] * 8, np.float64).tobytes())


def ref_exploded(fd):
    tree = {
        'data': np.arange(8)
    }

    asdf.AsdfFile(tree).write_to(fd, all_array_storage='external')


def ref_compressed(fd):
    tree = {
        'zlib': np.arange(128),
        'bzp2': np.arange(128)
    }

    ff = asdf.AsdfFile(tree)
    ff.set_array_compression(tree['zlib'], 'zlib')
    ff.set_array_compression(tree['bzp2'], 'bzp2')
    ff.write_to(fd)


def generate(version):
    outdir = os.path.join(os.path.dirname(__file__), '..', version)
    if not os.path.exists(outdir):
        os.makedirs(outdir)

    for name, func in globals().copy().items():
        if not name.startswith("ref_"):
            continue

        name = name[4:]
        filename = os.path.join(outdir, name)
        func(filename + ".asdf")
        with asdf.open(filename + ".asdf") as af:
            af.resolve_references()
            af.write_to(filename + ".yaml", all_array_storage="inline")


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser(
        "generate",
        description="Regenerate the ASDF reference files")
    parser.add_argument(
        "version", type=str, help="The ASDF version")
    args = parser.parse_args()

    with asdf.config_context() as cfg:
        cfg.default_version = args.version
        generate(args.version)



import re

import pytest
from common import (
    VALID_FILE_FORMAT_VERSIONS,
    VALID_YAML_VERSIONS,
    VERSION_MAP_PATHS,
    assert_yaml_header_and_footer,
    is_deprecated,
    load_yaml,
    path_to_id,
    split_id,
)
from packaging.version import Version

VALID_FILENAME_RE = re.compile(r"version_map-[0-9]+\.[0-9]+\.[0-9]+\.yaml")

SORTED_PATHS = sorted(VERSION_MAP_PATHS, key=lambda p: Version(split_id(path_to_id(p))[1]))

LATEST_PATH = SORTED_PATHS[-1]


@pytest.mark.parametrize("path", VERSION_MAP_PATHS)
def test_version_map(path):
    assert VALID_FILENAME_RE.match(path.name) is not None, f"{path.name} is an invalid version map filename"

    assert_yaml_header_and_footer(path)

    vm = load_yaml(path)

    assert set(vm.keys()) == {"FILE_FORMAT", "YAML_VERSION", "tags"}

    assert vm["FILE_FORMAT"] in VALID_FILE_FORMAT_VERSIONS
    assert vm["YAML_VERSION"] in VALID_YAML_VERSIONS

    assert len(vm["tags"].keys()) == len(set(vm["tags"].keys())), f"{path.name} contains duplicate tags"

    sorted_tags = sorted(list(vm["tags"].keys()))
    if sorted_tags != list(vm["tags"].keys()):
        sorted_list = "\n".join([f"""{tag}: {vm["tags"][tag]}""" for tag in sorted_tags])
        message = f"{path.name} tag list is not sorted.  Try this order instead:\n{sorted_list}"
        assert False, message


@pytest.mark.parametrize("path, previous_path", zip(SORTED_PATHS[1:], SORTED_PATHS[0:-1]))
def test_version_map_tags_retained(path, previous_path):
    """
    Confirm that non-deprecated tags were not lost between
    successive version maps.
    """
    vm = load_yaml(path)
    prev_vm = load_yaml(previous_path)

    expected_tags = {t for t in prev_vm["tags"].keys() if not is_deprecated(t)}
    tags = set(vm["tags"].keys())
    if not expected_tags.issubset(tags):
        missing_tags = expected_tags - tags
        assert False, (
            f"{path.name} is missing schemas that were present in the "
            "previous version.  If this was intentional, update the deprecation "
            "list in tests/common.py, otherwise add the following missing schemas: "
            f"""{", ".join(missing_tags)}"""
        )
